import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

public class RecipeGraph {
    
    private int size;
    private HashMap<String, Integer> indexer;
    private ArrayList<ArrayList<Integer>> graph;
    private ArrayList<String> labels;
    private ArrayList<String> types;
    
    public static final String ACTION_COLOR = "azure2";
    public static final String RAW_INGREDIENT_COLOR = "bisque1";
    public static final String INTERMEDIATE_INGREDIENT_COLOR = "lightgoldenrod1";
    
    public RecipeGraph() {
        size = 0;
        indexer = new HashMap<String, Integer>();
        graph = new ArrayList<ArrayList<Integer>>();
        labels = new ArrayList<String>();
        types = new ArrayList<String>();
    }
    
    // Add node u to the graph if it has not already been added
    public void addNode(String u, String label, String type) {
        if (indexer.containsKey(u))
            return;
        indexer.put(u, size);
        size++;
        graph.add(new ArrayList<Integer>());
        labels.add(label);
        types.add(type);
    }
    
    public void addNode(String u) {
        addNode(u, "NULL_LABEL", "NULL_TYPE");
    }
    
    // Add an edge from u to v. Requires that u and v already be in the graph
    public void addEdge(String u, String v) {
        if (!indexer.containsKey(u) || !indexer.containsKey(v))
            return;
        ArrayList<Integer> edges = graph.get(indexer.get(u));
        edges.add(indexer.get(v));
    }
    
    public void setType(String u, String type) {
        if (!indexer.containsKey(u))
            return;
        types.set(indexer.get(u), type);
    }
    
    public ArrayList<ArrayList<Integer>> getGraph() {
        return graph;
    }
    
    public ArrayList<String> getLabels() {
        return labels;
    }
    
    public ArrayList<String> getTypes() {
        return types;
    }
    
    public int getSize() {
        return size;
    }
    
    public void removeIntermediates() {
        ArrayList<ArrayList<Integer>> newGraph = new ArrayList<ArrayList<Integer>>();
        ArrayList<String> newLabels = new ArrayList<String>();
        ArrayList<String> newTypes = new ArrayList<String>();
        int[] lookup = new int[size];
        Arrays.fill(lookup, -1);
        int n = 0;
        
        for (int i = 0; i < size; i++) {
            if (types.get(i).equals("Action") || types.get(i).equals("RawIngredient")) {
                lookup[i] = n;
                n++;
                newGraph.add(new ArrayList<Integer>());
                newLabels.add(labels.get(i));
                newTypes.add(types.get(i));
            }
        }
        
        for (int i = 0; i < size; i++) {
            if (lookup[i] != -1) {
                ArrayList<Integer> edges = graph.get(i);
                for (int j : edges) {
                    if (lookup[j] == -1) {
                        for (int k : graph.get(j)) {
                            newGraph.get(lookup[i]).add(lookup[k]);
                        }
                    } else {
                        newGraph.get(lookup[i]).add(lookup[j]);
                    }
                }
            }
        }
        
        this.graph = newGraph;
        this.labels = newLabels;
        this.types = newTypes;
        this.size = n;
    }
    
    public static RecipeGraph makeIntersection(RecipeGraph g1, RecipeGraph g2, List<int[]> edits) {
        RecipeGraph intersection = new RecipeGraph();
        for (int[] e : edits) {
            if (e[0] != -1 && e[1] != -1) {
                intersection.addNode("S"+e[0]+e[1], g1.getLabels().get(e[0]), g1.getTypes().get(e[0]));
            }
        }
        
        for (int i=0; i < edits.size(); i++) {
            int [] source = edits.get(i);
            String s = "";
            if (source[0] != -1 && source[1] != -1) {
                s = "S"+source[0]+source[1];
                for (int j = 0; j < edits.size(); j++) {
                    int [] dest = edits.get(j);
                    if (dest[0] != -1 && g1.getGraph().get(source[0]).contains(dest[0])
                            && dest[1] != -1 && g2.getGraph().get(source[1]).contains(dest[1])) {
                        intersection.addEdge(s, "S"+dest[0]+dest[1]);
                    }
                }
            }
        }
        
        return intersection;
        
    }
    
    public static RecipeGraph makeUnion(RecipeGraph g1, RecipeGraph g2, List<int[]> edits) {
        
        RecipeGraph union = new RecipeGraph();
        for (int[] e : edits) {
            if (e[0] == -1) {
                union.addNode("G2"+e[1], "G2 " + g2.getLabels().get(e[1]), "Action");
            } else if (e[1] == -1) {
                union.addNode("G1"+e[0], "G1 " + g1.getLabels().get(e[0]), "Action");
            } else {
                // this will not work in the general case
                union.addNode("S"+e[0]+e[1], "S " + g1.getLabels().get(e[0]), "Action");
            }
        }
        
        for (int i=0; i < edits.size(); i++) {
            int [] source = edits.get(i);
            String s = "";
            if (source[0] == -1) {
                s = "G2"+source[1];
                for (int j = 0; j < edits.size(); j++) {
                    int [] dest = edits.get(j);
                    if (dest[1] != -1 && g2.getGraph().get(source[1]).contains(dest[1])) {
                        if (dest[0] == -1) {
                            union.addEdge(s, "G2"+dest[1]);
                        } else {
                            union.addEdge(s, "S"+dest[0]+dest[1]);
                        }
                    }
                }
            } else if (source[1] == -1) {
                s = "G1"+source[0];
                for (int j = 0; j < edits.size(); j++) {
                    int [] dest = edits.get(j);
                    if (dest[0] != -1 && g1.getGraph().get(source[0]).contains(dest[0])) {
                        if (dest[1] == -1) {
                            union.addEdge(s, "G1"+dest[0]);
                        } else {
                            union.addEdge(s, "S"+dest[0]+dest[1]);
                        }
                    }
                }
            } else {
                s = "S"+source[0]+source[1];
                for (int j = 0; j < edits.size(); j++) {
                    int [] dest = edits.get(j);
                    if (dest[0] != -1 && g1.getGraph().get(source[0]).contains(dest[0])) {
                        if (dest[1] == -1) {
                            union.addEdge(s, "G1"+dest[0]);
                        } else {
                            union.addEdge(s, "S"+dest[0]+dest[1]);
                        }
                    } else if (dest[1] != -1 && g2.getGraph().get(source[1]).contains(dest[1])) {
                        if (dest[0] == -1) {
                            union.addEdge(s, "G2"+dest[1]);
                        } else {
                            union.addEdge(s, "S"+dest[0]+dest[1]);
                        }
                    }
                }
            }
            
            
            
            
        }
        
        
        
        return union;
    }
    
    public String toDotFormat() {
        StringBuilder s = new StringBuilder();
        s.append("digraph G {\n");
        for (int nodeID = 0; nodeID < graph.size(); nodeID++) {
            s.append(nodeID);
            s.append(" [label=\"" + labels.get(nodeID) + "\", ");
            s.append("style=filled, ");
            String color = ACTION_COLOR;
            if (types.get(nodeID).equals("Ingredient")) {
                color = INTERMEDIATE_INGREDIENT_COLOR;
            }
            if (types.get(nodeID).equals("RawIngredient")) {
                color = RAW_INGREDIENT_COLOR;
            }
            s.append("fillcolor=" + color);
            s.append("];\n");
        }
        for (int u = 0; u < graph.size(); u++) {
            ArrayList<Integer> edges = graph.get(u);
            for (int v : edges) {
                s.append(u + " -> " + v + ";\n");
            }
        }
        s.append("}");
        return s.toString();
    }
    
    //
    //
    //
    //
    //
    
    static class GraphEditDistanceProblem {
        RecipeGraph startGraph;
        RecipeGraph endGraph;
        
        public GraphEditDistanceProblem(RecipeGraph rg1, RecipeGraph rg2) {
            startGraph = rg1;
            endGraph = rg2;
        }
        
        public List<int []> getStartState() {
            return new ArrayList<int[]>();
        }
        
        public boolean isCompleteEditPath(List<int []> path) {
            if(path.size() < startGraph.getSize())
                return false;
            int[] startGraphNodes = new int[startGraph.getSize()];
            int[] endGraphNodes = new int[endGraph.getSize()];
            for (int[] editOperation : path) {
                int startNode = editOperation[0];
                int endNode = editOperation[1];
                if (startNode >= 0) startGraphNodes[startNode] = 1;
                if (endNode >= 0) endGraphNodes[endNode] = 1;
            }
            for (int node : startGraphNodes)
                if (node == 0) return false;
            for (int node : endGraphNodes)
                if (node == 0) return false;
            return true;
        }
        
        public List<Pair<List<int[]>, Double>> getSuccessors(List<int []> path) {
            List<Pair<List<int[]>, Double>> successors = new ArrayList<Pair<List<int[]>,Double>>();
            int k = path.size();
            if (k < startGraph.getSize()) {
                int[] destNodes = new int[endGraph.getSize()];
                HashSet<Integer> deletedNodes = new HashSet<Integer>();
                HashMap<Integer, Integer> mappedNodes = new HashMap<Integer, Integer>();
                for (int[] edit : path) {
                    int editSrc = edit[0];
                    int editDest = edit[1];
                    if (editDest == -1) {
                        deletedNodes.add(editSrc);
                    } else {
                        mappedNodes.put(editSrc, editDest);
                        destNodes[editDest] = 1;
                    }
                }
                
                // get all the successors, for each one compute the step cost
                for (int i = 0; i < destNodes.length; i++) {
                    int inPath = destNodes[i];
                    boolean sameType = startGraph.types.get(k).equals(endGraph.types.get(i));
                    if (inPath == 0 && sameType) {
                        List<int[]> successorPath = new ArrayList<int[]>();
                        successorPath.addAll(path);
                        int[] newEdit = {k, i};
                        successorPath.add(newEdit);
                        
                        double cost = 0;
                        cost += nodeRelabelCost(startGraph.labels.get(k), endGraph.labels.get(i));
                        for (int originalNode : mappedNodes.keySet()) {
                            int mappedNode = mappedNodes.get(originalNode);
                            boolean outgoingOriginalEdge = startGraph.getGraph().get(k).contains(originalNode);
                            boolean outgoingMappedEdge = endGraph.getGraph().get(i).contains(mappedNode);
                            if (outgoingMappedEdge != outgoingOriginalEdge) cost += edgeInsertDeleteCost();
                            boolean incomingOriginalEdge = startGraph.getGraph().get(originalNode).contains(k);
                            boolean incomingMappedEdge = endGraph.getGraph().get(mappedNode).contains(i);
                            if (incomingOriginalEdge != incomingMappedEdge) cost += edgeInsertDeleteCost();
                        }
                        successors.add(Pair.makePair(successorPath, cost));
                    }
                }
                
                // insert the node deletion
                List<int[]> deletionSuccessorPath = new ArrayList<int[]>();
                deletionSuccessorPath.addAll(path);
                int[] newEdit = {k, -1};
                deletionSuccessorPath.add(newEdit);
                
                double cost = nodeInsertDeleteCost();
                for (int i = 0; i < startGraph.getSize(); i++) {
                    if (!deletedNodes.contains(i) && i!=k) {
                        if (startGraph.getGraph().get(k).contains(i))
                            cost += edgeInsertDeleteCost();
                        if (startGraph.getGraph().get(i).contains(k))
                            cost += edgeInsertDeleteCost();
                    }
                }
                
                successors.add(Pair.makePair(deletionSuccessorPath, cost));
            } else {
                successors.addAll(getInsertionSuccessors(path));
            }
            return successors;
        }
        
        private List<Pair<List<int[]>, Double>> getInsertionSuccessors(List<int []> path) {
            // There is only one insertion successor
            int[] endGraphNodes = new int[endGraph.getSize()];
            for (int[] editOperation : path) {
                int endNode = editOperation[1];
                if (endNode >= 0) endGraphNodes[endNode] = 1;
            }
            
            List<Pair<List<int[]>, Double>> successors = new ArrayList<Pair<List<int[]>,Double>>();
            List<int[]> insertionSuccessor = new ArrayList<int[]>();
            insertionSuccessor.addAll(path);
            double cost = 0;
            for (int i = 0; i < endGraphNodes.length; i++) {
                int endNodeInPath = endGraphNodes[i];
                if (endNodeInPath == 0) {
                    cost += edgeInsertDeleteCost()*endGraph.getGraph().get(i).size();
                    cost += nodeInsertDeleteCost();
                    int[] newEdit = {-1, i};
                    insertionSuccessor.add(newEdit);
                } else {
                    List<Integer> adjacentNodes = endGraph.getGraph().get(i);
                    for (int node : adjacentNodes) {
                        if (endGraphNodes[node] == 0) cost += edgeInsertDeleteCost();
                    }
                }
            }
            successors.add(Pair.makePair(insertionSuccessor, cost));
            return successors;
        }
        
        public double getCostOfActions() {
            throw new UnsupportedOperationException();
        }
        
        public double edgeInsertDeleteCost() {
            return 1.0;
        }
        
        public double nodeInsertDeleteCost() {
            return 1.0;
        }
        
        public double nodeRelabelCost(String u, String v) {
            if (u.equals(v)) {
                return 0.0;
            }
            return Double.POSITIVE_INFINITY;
        }
        
    }
    
    public static Pair<List<int[]>, Double> editDistance(RecipeGraph rg1, RecipeGraph rg2) {
        GraphEditDistanceProblem problem = new GraphEditDistanceProblem(rg1, rg2);
        Pair<List<int[]>, Double> editPath = aStarSearch(problem);
        return editPath;
    }
    
    public static Pair<List<int[]>, Double> aStarSearch(GraphEditDistanceProblem problem) {
        FastPriorityQueue<Pair<List<int[]>, Double>> fringe = new FastPriorityQueue<Pair<List<int[]>, Double>>();
        List<int[]> currentPath = problem.getStartState();
        double currentCost = 0;
        fringe.setPriority(Pair.makePair(currentPath, currentCost), 0);
        while(true) {
            if(fringe.isEmpty()) return null;
            Pair<List<int[]>, Double> pair = fringe.removeFirst();
            currentPath = pair.getFirst();
            currentCost = pair.getSecond();
            if (problem.isCompleteEditPath(currentPath))
                return pair;
            //
            List<Pair<List<int[]>, Double>> successors = problem.getSuccessors(currentPath);
            for (Pair<List<int[]>, Double> successor : successors) {
                double stepCost = successor.getSecond();
                fringe.setPriority(Pair.makePair(successor.getFirst(), currentCost + stepCost), -(currentCost + stepCost));
            }
        }
    }
    
    public static RecipeGraph generateRandomGraph(int size) {
        RecipeGraph g = new RecipeGraph();
        for (int i = 0; i < size; i++) {
            g.addNode(String.valueOf(i));
        }
        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                if (i != j) {
                    double rand = Math.random();
                    if (rand > 0.7) {
                        g.addEdge(String.valueOf(i), String.valueOf(j));
                    }
                }
            }
        }
        return g;
    }
    
    public static void main(String[] args) {
        RecipeGraph g1 = generateRandomGraph(3);
        RecipeGraph g2 = generateRandomGraph(3);
        long startTime = System.currentTimeMillis();
        Pair<List<int[]>, Double> pair = editDistance(g1, g2);
        long endTime = System.currentTimeMillis();
        System.out.println(endTime - startTime);
        System.out.println(pair.getSecond());
    }
}
