package vis;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;

public class Graph {
    public ArrayList<Vertex> vertices;
    public double[][] distances;
    
    public Graph () {
        this.vertices = new ArrayList<Vertex>();
    }
    
    public void addVertex(Vertex v) {
        vertices.add(v);
    }
    
    public void addVertex(String label) {
        Vertex v = new Vertex(label);
        vertices.add(v);
    }
    
    public void addVertices(String[] labels) {
        for (String label: labels) {
            addVertex(label);
        }
    }
    
    public void addEdge(String vLabel, String uLabel, double w) {
        Vertex v = null;
        Vertex u = null;
        for (int i = 0; i < vertices.size(); i++) {
            Vertex x = vertices.get(i);
            if (x.label.equals(vLabel)) {
                v = x;
            }
            if (x.label.equals(uLabel)) {
                u = x;
            }
        }
        addEdge(v,u,w);
    }
    
    public void addEdge(Vertex v, Vertex u, double w) {
        v.addEdge(new Edge(u, w).setDirected());
        u.addEdge(new Edge(v, w));
    }
    
    public void computeDistances() {
        int n = vertices.size(); 
        distances = new double[n][n];
        for (int i = 0; i < n; i++) {
            // compute the shortest paths from this src to everything else
            double[] dist = new double[n];
            Arrays.fill(dist, Double.POSITIVE_INFINITY);
            dist[i] = 0;
            boolean[] Q = new boolean[n];
            Arrays.fill(Q, true);
            int c = 0;
            while (c < n) {
                double min = Double.POSITIVE_INFINITY;
                int minIndex = -1;
                for (int j = 0; j < n; j++) {
                    if (Q[j] && dist[j] < min) {
                        min = dist[j];
                        minIndex = j;
                    }
                }
                if (min == Double.POSITIVE_INFINITY) break;
                Q[minIndex] = false;
                c++;
                for (Edge e : vertices.get(minIndex).edges) {
                    int neighbor = vertices.indexOf(e.d); 
                    if (Q[neighbor]) {
                        double newDist = min + e.weight;
                        if (newDist < dist[neighbor]) {
                            dist[neighbor] = newDist;
                        }
                    }
                }
            }
            distances[i] = dist;
        }
    }
    
    public void computeNodePositions() {
        int numIter = 100000;
        int n = vertices.size();
        double eps = 0.01; // step size
        for (int iter = 0; iter < numIter; iter++) {
            // compute the derivative for each node then move all nodes
            for (int i = 0; i < n; i++) {
                Vertex v_i = vertices.get(i);
                double dx = 0.0;
                double dy = 0.0;
                for (int j = 0; j < n; j++) {
                    if (i != j ) {
                        if (distances[i][j] != 0) {
                            Vertex v_j = vertices.get(j);
                            double x_diff = v_i.x - v_j.x;
                            double y_diff = v_i.y - v_j.y;
                            double norm = Math.sqrt(x_diff*x_diff + y_diff*y_diff);
                            double w = 1.0 / (distances[i][j]*distances[i][j]); // don't actually need to compute this every time
                            double c = w*2*(norm - distances[i][j]) / norm;
                            dx += c * x_diff;
                            dy += c * y_diff;
                        }
                    }
                }
                v_i.dx = dx * eps;
                v_i.dy = dy * eps;
            }
            for (int i = 0; i < n; i++) {
                Vertex v_i = vertices.get(i);
                v_i.x = v_i.x - v_i.dx;
                v_i.y = v_i.y - v_i.dy;
            }
        }
    }
    
    public void printGraph() {
        for (int i = 0; i < vertices.size(); i++) {
            String s = vertices.get(i).label + ": ";
            for (Edge e : vertices.get(i).edges) {
                s += e.d.label + ", ";
            }
            System.out.println(s);
        }
    }
    
    public void printDistances() {
        int n = vertices.size();
        for (int i = 0; i < n; i++) {
            System.out.println();
            for (int j = 0; j < n; j++) {
                System.out.print(distances[i][j] + " ");
            }
        }
    }
    
    public String toDotFormat() {
        StringBuilder s = new StringBuilder();
        s.append("graph G {\n");
        for (int nodeID = 0; nodeID < vertices.size(); nodeID++) {
            Vertex u = vertices.get(nodeID);
            s.append(nodeID);
            s.append(" [label=\"" + u.label + "\", ");
            //s.append(" [label=\"" + vertices.get(nodeID).label + "\"");
            s.append("pos=\"" + u.x + "," + u.y + "!\"");
            s.append("];\n");
        }
        for (int u = 0; u < vertices.size(); u++) {
            for (Edge e : vertices.get(u).edges) {
                int v = vertices.indexOf(e.d);
                if (v != -1 && u < v) s.append(u + " -- " + v + ";\n");
            }
        }
        s.append("}");
        return s.toString();
    }
    
    public String toDotFormatDirected() {
        StringBuilder s = new StringBuilder();
        s.append("digraph G {\n");
        for (int nodeID = 0; nodeID < vertices.size(); nodeID++) {
            Vertex u = vertices.get(nodeID);
            s.append(nodeID);
            s.append(" [label=\"" + u.label + "\", ");
            s.append("pos=\"" + u.x + "," + u.y + "!\", ");
            s.append("style=filled, ");
            s.append("fillcolor=" + u.color);
            s.append("];\n");
        }
        for (int u = 0; u < vertices.size(); u++) {
            for (Edge e : vertices.get(u).edges) {
                int v = vertices.indexOf(e.d);
                if (v != -1 && e.directed) s.append(u + " -> " + v + ";\n");
            }
        }
        s.append("}");
        return s.toString();
    }
    
    public void output(String fname, boolean digraph){
        String dotFormatted = toDotFormat();
        if (digraph) dotFormatted = toDotFormatDirected();
        try {
            BufferedWriter out = new BufferedWriter(new FileWriter("C:\\Users\\Matt\\Desktop\\graphs\\"+fname));
            out.write(dotFormatted);
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    public static Graph getRandomGraph(String[] labels, double p) {
        Graph g = new Graph();
        
        // add the vertices
        for (int i = 0; i < labels.length; i++) {
            Vertex v = new Vertex(labels[i]);
            g.addVertex(v);
        }
        
        // add some random edges
        for (int i = 0; i < g.vertices.size(); i++) {
            for (int j = i+1; j < g.vertices.size(); j++) {
                double r = Math.random();
                if (r < p) {
                    Edge e1 = new Edge(g.vertices.get(j), 1.0);
                    g.vertices.get(i).addEdge(e1);
                    Edge e2 = new Edge(g.vertices.get(i), 1.0);
                    g.vertices.get(j).addEdge(e2);
                }
            }
        }
        
        return g;
    }
    
    public static void main(String[] args) {
        
        String[] labels = {"1", "2", "3", "4","5"};
        Graph g = getRandomGraph(labels, 0.5);
        g.printGraph();
        System.out.println();
        g.computeDistances();
        g.printDistances();
        
        g.output("random", false);
        
    }
}
